import re


def parse_fgdc_metadata(md_text):
    line_pattern = r'([\s]*)([^:]+)*(\:?)[\s]*(.*)'

    def _parse(lines, group_indent=0):
        result = {}
        is_multiline_text = False
        result_text = None
        while len(lines) > 0:
            # Peek at the next line to see if we're done this group
            line = lines[0]
            # Replace stray carriage returns
            line = line.replace('^M', '')
            # Replace tabs with 4 spaces
            line = line.replace('\t', '    ')

            if not line.strip():
                lines.pop(0)
                continue

            m = re.search(line_pattern, line)
            if not m:
                raise Exception('Could not parses line:\n{}\n'.format(line))

            this_indent = len(m.group(1))

            if this_indent < group_indent:
                break

            # Remove this line from the top of the stack and
            # process it.

            lines.pop(0)

            # If we're collecting multi-line text values,
            # just add the line to the result text.
            if is_multiline_text:
                result_text += ' {}'.format(line.strip())
                continue

            this_key = m.group(2)
            key_flag = m.group(3) != '' and ' ' not in this_key
            this_value = m.group(4)

            if key_flag and this_value != '':
                result[this_key] = this_value.strip()
            else:
                if key_flag:
                    result[this_key] = _parse(lines, group_indent + 2)
                else:
                    is_multiline_text = True
                    result_text = line.strip()

        if is_multiline_text:
            return result_text
        else:
            return result

    return _parse(md_text.split('\n'), group_indent=0)['Metadata']



import pystac
from pystac.extensions.eo import Band

NAIP_ID = 'usda-naip'
NAIP_TITLE = "NAIP: National Agriculture Imagery Program"
NAIP_DESCRIPTION = """
The National Agriculture Imagery Program (NAIP) acquires aerial imagery
during the agricultural growing seasons in the continental U.S.

NAIP projects are contracted each year based upon available funding and the
FSA imagery acquisition cycle. Beginning in 2003, NAIP was acquired on
a 5-year cycle. 2008 was a transition year, and a three-year cycle began
in 2009.

NAIP imagery is acquired at a one-meter ground sample distance (GSD) with a
horizontal accuracy that matches within six meters of photo-identifiable
ground control points, which are used during image inspection.

Older images were collected using 3 bands (Red, Green, and Blue: RGB), but
newer imagery is usually collected with an additional near-infrared band
(RGBN).
""".strip('\n')

NAIP_LICENSE = 'PDDL-1.0'

USDA_PROVIDER = pystac.Provider(
    name='USDA Farm Service Agency',
    url=('https://www.fsa.usda.gov/programs-and-services/aerial-photography'
         '/imagery-programs/naip-imagery/'),
    roles=['producer', 'licensor'])

NAIP_BANDS = [
    Band.create(name="Red", common_name='red'),
    Band.create(name="Green", common_name='green'),
    Band.create(name="Blue", common_name='blue'),
    Band.create(name="NIR", common_name='nir', description="near-infrared")
]



# flake8: noqa

from stactools.naip.stac import create_item

import stactools.core

stactools.core.use_fsspec()


def register_plugin(registry):
    # Register subcommands

    from stactools.naip import commands

    registry.register_subcommand(commands.create_naip_command)



import os
from imp import load_source
from setuptools import setup, find_namespace_packages
from glob import glob
import io

name = 'stactools_naip'
description = ("Subpackage for working with NAIP  data in stactools, "
               "a command line tool and Python library for working with STAC.")

__version__ = load_source(
    'stactools.naip.version',
    os.path.join(os.path.dirname(__file__),
                 'stactools/naip/version.py')).__version__

from os.path import (basename, splitext)

here = os.path.abspath(os.path.dirname(__file__))

# get the dependencies and installs
with io.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    install_requires = [line.split(' ')[0] for line in f.read().split('\n')]

# Add stactools subpackage dependencies
install_requires.extend(['stactools_core=={}'.format(__version__)])

with open(os.path.join(here, 'README.md')) as readme_file:
    readme = readme_file.read()

setup(name=name,
      description=description,
      version=__version__,
      long_description=readme,
      long_description_content_type="text/markdown",
      author="stac-utils",
      author_email='stac@radiant.earth',
      url='https://github.com/stac-utils/stactools.git',
      packages=find_namespace_packages(),
      py_modules=[
          splitext(basename(path))[0] for path in glob('stactools/*.py')
      ],
      include_package_data=False,
      install_requires=install_requires,
      license="Apache Software License 2.0",
      keywords=[
          'stactools', 'psytac', 'naip', 'imagery', 'raster', 'catalog',
          'STAC'
      ])



import logging
import json
import os

import click
import pystac

from stactools.naip.stac import create_item

logger = logging.getLogger(__name__)


def create_naip_command(cli):
    """Creates a command group for commands working with
    NAIP imagery.
    """
    @cli.group('naip',
               short_help=("Commands for working with "
                           "NAIP imagery."))
    def naip():
        pass

    @naip.command('create-item',
                  short_help='Create a STAC Item from NAIP imagery data')
    @click.argument('state')
    @click.argument('year')
    @click.argument('cog_href')
    @click.argument('dst')
    @click.option('-f', '--fgdc', help='HREF to FGDC metadata.')
    @click.option('-t',
                  '--thumbnail',
                  help='HREF to the thumbnail for this NAIP tile')
    @click.option(
        '-p',
        '--providers',
        help='Path to JSON file containing array of additional providers')
    def create_item_command(state, year, cog_href, dst, fgdc, thumbnail,
                            providers):
        """Creates a STAC Item based on metadata from a NAIP tile.

        STATE is the state this NAIP tile belongs to.
        COG_HREF is the href to the COG that is the NAIP tile.
        FGDC_HREF is href to the text metadata file in the NAIP fgdc format.
        DST is directory that a STAC Item JSON file will be created
        in.
        """
        additional_providers = None
        if providers is not None:
            with open(providers) as f:
                additional_providers = [
                    pystac.Provider.from_dict(d) for d in json.load(f)
                ]

        item = create_item(state,
                           year,
                           cog_href,
                           fgdc_metadata_href=fgdc,
                           thumbnail_href=thumbnail,
                           additional_providers=additional_providers)

        item_path = os.path.join(dst, '{}.json'.format(item.id))
        item.set_self_href(item_path)

        item.save_object()

    return naip



import os
from typing import List, Optional

import dateutil.parser
import pystac
from pystac.utils import str_to_datetime
from shapely.geometry import shape, box, mapping
import rasterio as rio

from stactools.core.projection import reproject_geom
from stactools.naip import constants
from stactools.naip.utils import parse_fgdc_metadata


def naip_item_id(state, resource_name):
    """Generates a STAC Item ID based on the state and the "Resource Description"
    contained in the FGDC metadata.

    Args:
        state (str): The two-letter state code for the state this belongs to.
        resource_name (str): The resource name, e.g. m_3008501_ne_16_1_20110815_20111017.tif

    Returns:
        str: The STAC ID to use for this scene.
    """

    return '{}_{}'.format(state, os.path.splitext(resource_name)[0])


def create_collection(seasons: List[int]) -> pystac.Collection:
    """Creates a STAC COllection for NAIP data.

    Args:
        seasons (List[int]): List of years that represent the NAIP seasons
            this collection represents.
    """
    extent = pystac.Extent(
        pystac.SpatialExtent(bboxes=[[-124.784, 24.744, -66.951, 49.346]]),
        pystac.TemporalExtent(intervals=[[
            pystac.utils.str_to_datetime(f"{min(seasons)}-01-01T00:00:00Z"),
            pystac.utils.str_to_datetime(f"{max(seasons)}-01-01T00:00:00Z")
        ]]))

    collection = pystac.Collection(
        id=constants.NAIP_ID,
        description=constants.NAIP_DESCRIPTION,
        title=constants.NAIP_TITLE,
        license=constants.NAIP_LICENSE,
        providers=[constants.USDA_PROVIDER],
        extent=extent,
        stac_extensions=['item-assets'],
        extra_fields={
            'item_assets': {
                'image': {
                    "eo:bands": [b.properties for b in constants.NAIP_BANDS],
                    "roles": ["data"],
                    "title": "RGBIR COG tile",
                    "type": pystac.MediaType.COG
                },
            }
        })

    return collection


def create_item(state,
                year,
                cog_href,
                fgdc_metadata_href: Optional[str],
                thumbnail_href=None,
                additional_providers=None):
    """Creates a STAC Item from NAIP data.

    Args:
        state (str): The 2-letter state code for the state this item belongs to.
        year (str): The NAIP year.
        fgdc_metadata_href (str): The href to the FGDC metadata
            for this NAIP scene. Optional, a some NAIP scenes to not have this
            (e.g. 2010)
        cog_href (str): The href to the image as a COG. This needs
            to be an HREF that rasterio is able to open.
        thumbnail_href (str): Optional href for a thumbnail for this scene.
        additional_providers(List[pystac.Provider]): Optional list of additional
            providers to the USDA that will be included on this Item.

    This function will read the metadata file for information to place in
    the STAC item.

    Returns:
        pystac.Item: A STAC Item representing this NAIP scene.
    """

    with rio.open(cog_href) as ds:
        gsd = ds.res[0]
        epsg = int(ds.crs.to_authority()[1])
        image_shape = list(ds.shape)
        original_bbox = list(ds.bounds)
        transform = list(ds.transform)
        geom = reproject_geom(ds.crs,
                              'epsg:4326',
                              mapping(box(*ds.bounds)),
                              precision=6)

    if fgdc_metadata_href is not None:
        fgdc_metadata_text = pystac.STAC_IO.read_text(fgdc_metadata_href)
        fgdc = parse_fgdc_metadata(fgdc_metadata_text)
    else:
        fgdc = {}

    if 'Distribution_Information' in fgdc:
        resource_desc = fgdc['Distribution_Information'][
            'Resource_Description']
    else:
        resource_desc = os.path.basename(cog_href)
    item_id = naip_item_id(state, resource_desc)

    bounds = list(shape(geom).bounds)

    if any(fgdc):
        dt = str_to_datetime(
            fgdc['Identification_Information']['Time_Period_of_Content']
            ['Time_Period_Information']['Single_Date/Time']['Calendar_Date'])
    else:
        fname = os.path.splitext(os.path.basename(cog_href))[0]
        fname_date = fname.split('_')[5]
        dt = dateutil.parser.isoparse(fname_date)

    properties = {'naip:state': state, 'naip:year': year}

    item = pystac.Item(id=item_id,
                       geometry=geom,
                       bbox=bounds,
                       datetime=dt,
                       properties=properties)

    # Common metadata
    item.common_metadata.providers = [constants.USDA_PROVIDER]
    if additional_providers is not None:
        item.common_metadata.providers.extend(additional_providers)
    item.common_metadata.gsd = gsd

    # eo, for asset bands
    item.ext.enable('eo')

    # proj
    item.ext.enable('projection')
    item.ext.projection.epsg = epsg
    item.ext.projection.shape = image_shape
    item.ext.projection.bbox = original_bbox
    item.ext.projection.transform = transform

    # COG
    item.add_asset(
        'image',
        pystac.Asset(href=cog_href,
                     media_type=pystac.MediaType.COG,
                     roles=['data'],
                     title="RGBIR COG tile"))

    # Metadata
    if any(fgdc):
        item.add_asset(
            'metadata',
            pystac.Asset(href=fgdc_metadata_href,
                         media_type=pystac.MediaType.TEXT,
                         roles=['metadata'],
                         title='FGDC Metdata'))

    if thumbnail_href is not None:
        media_type = pystac.MediaType.JPEG
        if thumbnail_href.lower().endswith('png'):
            media_type = pystac.MediaType.PNG
        item.add_asset(
            'thumbnail',
            pystac.Asset(href=thumbnail_href,
                         media_type=media_type,
                         roles=['thumbnail'],
                         title='Thumbnail'))

    item.ext.eo.set_bands(constants.NAIP_BANDS, item.assets['image'])

    return item
